@@ -98,7 +98,7 @@ except ImportError:
 
 __pychecker__ = 'no-reimportself'
 
-__version__ = '2.0'
+__version__ = '2.1'
 
 PERM_READ = 0o400      # Read permission bit.
 PERM_WRITE = 0o200     # Write permission bit.
@@ -172,6 +172,7 @@ class FakeFile(object):
     self.name = name
     self.st_mode = st_mode
     self.contents = contents
+    self.epoch = 0
     self.st_ctime = int(time.time())
     self.st_atime = self.st_ctime
     self.st_mtime = self.st_ctime
@@ -220,6 +221,7 @@ class FakeFile(object):
     """
     self.contents = contents
     self.st_size = len(contents)
+    self.epoch += 1
 
   def SetSize(self, st_size):
     """Resizes file content, padding with nulls if new size exceeds the old.
@@ -243,6 +245,7 @@ class FakeFile(object):
     else:
       self.contents = '%s%s' % (self.contents, '\0' * (st_size - current_size))
     self.st_size = len(self.contents)
+    self.epoch += 1
 
   def SetATime(self, st_atime):
     """Set the self.st_atime attribute.
@@ -571,7 +574,10 @@ class FakeFilesystem(object):
       raise TypeError
     if not file_path:
       return False
-    file_path = self.ResolvePath(file_path)
+    try:
+      file_path = self.ResolvePath(file_path)
+    except IOError:
+      return False
     if file_path == self.root.name:
       return True
     path_components = self.GetPathComponents(file_path)
@@ -614,13 +620,20 @@ class FakeFilesystem(object):
 
     Raises:
       TypeError: if file_path is None
-      IOError: if file_path is ''
+      IOError: if file_path is '' or a part of the path doesn't exist
     """
 
     def _ComponentsToPath(component_folders):
       return '%s%s' % (self.path_separator,
                        self.path_separator.join(component_folders))
 
+    def _ValidRelativePath(file_path):
+      while file_path and '/..' in file_path:
+        file_path = file_path[:file_path.rfind('/..')]
+        if not self.Exists(self.NormalizePath(file_path)):
+          return False
+      return True
+
     def _FollowLink(link_path_components, link):
       """Follow a link w.r.t. a path resolved so far.
 
@@ -660,8 +673,9 @@ class FakeFilesystem(object):
     if file_path is None:
       # file.open(None) raises TypeError, so mimic that.
       raise TypeError('Expected file system path string, received None')
-    if not file_path:
-      # file.open('') raises IOError, so mimic that.
+    if not file_path or not _ValidRelativePath(file_path):
+      # file.open('') raises IOError, so mimic that, and validate that all
+      # parts of a relative path exist.
       raise IOError(errno.ENOENT,
                     'No such file or directory: \'%s\'' % file_path)
     file_path = self.NormalizePath(file_path)
@@ -1907,6 +1921,7 @@ class FakeFileOpen(object):
         self._read = read
         self._update = update
         self._closefd = closefd
+        self._file_epoch = file_object.epoch
         newline_arg = {} if binary else {'newline': newline}
         if file_object.contents:
           if update:
@@ -1962,6 +1977,7 @@ class FakeFileOpen(object):
       def flush(self):
         """Flush file contents to 'disk'."""
         self._file_object.SetContents(self._io.getvalue())
+        self._file_epoch = self._file_object.epoch
 
       def seek(self, offset, whence=0):
         """Move read/write pointer in 'file'."""
@@ -1987,6 +2003,17 @@ class FakeFileOpen(object):
           self._io.seek(write_seek)
         return self._read_seek
 
+      def _UpdateStringIO(self):
+        """Updates the StringIO with changes to the file object contents."""
+        if self._file_epoch == self._file_object.epoch:
+          return
+        whence = self._io.tell()
+        self._io.seek(0)
+        self._io.truncate()
+        self._io.write(self._file_object.contents)
+        self._io.seek(whence)
+        self._file_epoch = self._file_object.epoch
+
       def _ReadWrappers(self, name):
         """Wrap a StringIO attribute in a read wrapper.
 
@@ -2074,8 +2101,10 @@ class FakeFileOpen(object):
         if self._file_object.IsLargeFile():
           raise FakeLargeFileIoException(file_path)
 
+        if name.startswith('read'):
+          self._UpdateStringIO()
         if self._append:
-          if name in ['read', 'readline', 'readlines']:
+          if name.startswith('read'):
             return self._ReadWrappers(name)
           else:
             return self._OtherWrapper(name)
